@@ -51,6 +51,8 @@ pub struct ModifiedPage<'a, K: ?Sized, V: ?Sized, P: BTreePage<K, V>> {
     pub c1: P::Cursor,
     // Whether to skip `c1`'s first element.
     pub skip_first: bool,
+    // Whether the page we just modified is `self.l`.
+    pub mod_is_left: bool,
 }
 
 /// Represents the concatenation of a modified page and one of its
@@ -264,6 +266,11 @@ fn leaf_delete<
             r: 0,
             ins: None,
             ins2: None,
+            // The following (`mod_is_left`) is meaningless in this
+            // context, and isn't actually used: indeed, the only
+            // place where this field is used is when modifying the
+            // root, when `ins.is_some()`.
+            mod_is_left: true,
         },
         deleted,
     ))
@@ -370,7 +377,9 @@ fn handle_merge<
     k0v0: &Option<P::Saved>,
     incr_other: Option<CowPage>,
     incr_mid: Option<(&K, &V)>,
-    mod_is_left: bool,
+
+    mod_is_left: bool,    // The modified page in the `merge` is the left one.
+
     merge: Op<'a, T, K, V>,
     free: &mut [[u64; 2]; N_CURSORS],
 ) -> Result<ModifiedPage<'a, K, V, P>, T::Error> {
@@ -390,6 +399,7 @@ fn handle_merge<
             ins2: None,
             c1,
             skip_first: true,
+            mod_is_left,
         }
     };
 
@@ -488,7 +498,7 @@ fn handle_merge<
             // either (1) a rebalance replaces a key/value pair with a
             // larger one or (2) another split has happened in a page
             // below.
-            let split_key_is_k0 = if cursor.len() + 1 == p0 {
+            let split_key_is_k0 = if cursor.len() == p0 {
                 // In this case, ins2 comes after ins, since the
                 // replacement is in the right child of the deleted
                 // key.
@@ -501,10 +511,9 @@ fn handle_merge<
                 last_op.ins = Some((split_key, split_value));
                 last_op.skip_first = false;
 
-                // If this split isn't at the root, and the page
-                // modified in the last step is the one on the right
-                // of the current entry, move right one step before
-                // inserting the split key/value.
+                // If the page modified in the last step is the one on
+                // the right of the current entry, move right one step
+                // before inserting the split key/value.
                 //
                 // (remember that we popped the stack upon entering
                 // this function).
@@ -738,12 +747,25 @@ fn modify<'a, T: AllocPage, K: Storable + ?Sized, V: Storable + ?Sized, P: BTree
     if let Some((k, v)) = m.ins {
         // The following call might actually replace the first element
         // of `m.c1` if `m.skip_first`.
+        let mut c1 = m.c1.clone();
+        if !m.skip_first && !m.mod_is_left {
+            // This means that the page below just split, since we
+            // have to insert an extra entry on the root page.
+            //
+            // However, the extra entry is to be inserted (by
+            // `P::put`) *before* `c1`'s first element, which is
+            // incorrect since the page that split is the right child
+            // of `c1`'s first element. Therefore, we need to move
+            // `c1` one notch to the right.
+            assert!(m.ins2.is_none());
+            P::move_next(&mut c1);
+        }
         P::put(
             txn,
             m.page,
             m.mutable,
             m.skip_first,
-            &m.c1,
+            &c1,
             k,
             v,
             m.ins2,
